-- Time-stamp: <Wed Sep 29 2010 20:56:57 Stardate: Stardate: [-28]3758.94 hwloidl>
-- $Id: LUDecompMatrix.hs,v 1.1.2.1 2002/06/15 01:34:29 hwloidl Exp $
--
-- Doing an LU decomposition on a square matrix represented as an Array.
-- The full details of this program, including a mathematical derivation
-- from its matrix algebra specification, can be found in:
-- 
-- Hudak, P., "Para-functional Programming in Haskell",
--   chapter 5 in "Parallel Functional Languages and Compilers",
--   edited by Szymanski, B.K., ACM Press (New York) and Addison-Wesley
--   (Reading), 1991.
--
-- Written by Paul Hudak, March 1993
--
-- Variants (CPP vars):
--  MOD_LU ... use modular arithmetic in LU decomposition and 
--             fwd/bwd_substitution to solve
--             top level function:
--             homsolv :: (Integral a) => 
--                        a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
--  TRACE ... trace messages for debugging
-- -------------------------------------------------------------------------

module LUDecompMatrix(lu, det, 
                homsolv,
                -- DEBUGGING ONLY        
                mult_mat_vec, mult_mat_mat, mult_vec_vec, 
                fwd_subst, bwd_subst, split_lu, check_lu, do_lu,
               ) where

-- import Array
-- import Ratio 
import Random
--import Strategies
import ModArithm
import Matrix 

#if defined(STRATEGIES)
# if defined(EVAL_STRATEGIES)
import Control.DeepSeq
import Control.Parallel
import Control.Parallel.Strategies
# elif defined(SM_STRATEGIES)
import Control.DeepSeq
import Control.Parallel
import Control.Parallel.Strategies
# else
import Control.Parallel
import Control.Parallel.Strategies
# endif
#else
import Control.DeepSeq
#endif

-- Aux fcts

{-# SPECIALISE 
       sumOf :: Int -> (Int,Int) -> (Int -> Int) -> Int
  #-}
sumOf :: (Integral a, Enum b) => a -> (b,b) -> (b -> a) -> a
sumOf m (a,b) f = foldl (\ z x -> modSum m z x) 0 (map f [a..b])

{-
-- Hudak chapter
lu :: (Fractional a) => 
      Array (Int, Int) a -> Array (Int, Int) a
lu a = let 
         lua = array bds
               ( [ ((i,j), doL i j)  |  i<-[rl..rh], j<-[cl..i]   ] ++
                 [ ((i,j), doU i j)  |  i<-[rl..rh], j<-[i+1..ch] ] 
                 )
         doL i j = (a!(i,j) - sumOf (1,j-1) (\k -> lua!(i,k) * lua!(k,j)))
         doU i j 
           | lua!(i,i)==0 = error ("doU: 0 entry at (" ++ (show i) ++ "," ++ (show i) ++ ")")
           | otherwise = (a!(i,j) - sumOf (1,i-1) (\k -> lua!(i,k) * lua!(k,j))) 
                          / (lua!(i,i))
         bds@((rl,cl),(rh,ch)) = bounds a
       in 
       lua
-}

-- this uses modular arithmetic
{-# SPECIALISE 
  lu :: Int -> SqMatrix Int -> SqMatrix Int
  #-}
lu :: (Integral a) => 
      a -> SqMatrix a -> SqMatrix a
lu m a = 
  let 
    lua =  lolSqMatrix bds 
           [ [ {-(i,j),-} x | j<-[cl..ch], 
                              let x = if i<=j then doU i j else doL i j ]
             | i <-[rl..rh] ]
    -- i>j: (2.3.13)
    doL i j = modProd m 
                (modInv m (lua!!-(j,j)))
                (modDif m (a!!-(i,j))
                          (sumOf m (1,j-1) (\k -> (modProd m (lua!!-(i,k)) 
                                                             (lua!!-(k,j))))))
    -- i<=j: (2.3.12)
    doU i j = modDif m (a!!-(i,j)) 
                       (sumOf m (1,i-1) (\k -> modProd m (lua!!-(i,k)) 
                                                         (lua!!-(k,j)))) 
    bds@((rl,cl),(rh,ch)) = matBounds a
  in 
  lua

{-# SPECIALISE 
  fwd_subst :: Int -> SqMatrix Int -> Vector Int -> Vector Int
  #-}
fwd_subst :: (Integral a) => 
             a -> SqMatrix a -> Vector a -> Vector a
fwd_subst m lua b = 
  let
    y = vector y'
    y' = [ {-one,-} modQuot m (b!-1) (lua!!-(one,one)) ] ++ 
         [ {-i,-} modProd m (modInv m (lua!!-(i,i)))
                            (modDif m (b!-i) 
                              (sumOf m (one,i-1) (\j -> modProd m (lua!!-(i,j)) (y!-j))))
           | i <- [one+1..n] ]
             
    bds@(one,n) = vecBounds b
    bds'@((rl,cl),(rh,ch)) = matBounds lua
  in
    y

{-# SPECIALISE 
  bwd_subst :: Int -> SqMatrix Int -> Vector Int -> Vector Int
  #-}
bwd_subst :: (Integral a) => 
             a -> SqMatrix a -> Vector a -> Vector a
bwd_subst m lua y =
  let
    x = vector x'
    x' = reverse $
         [ {-n,-} modQuot m (y!-n) (lua!!-(n,n)) ] ++
         [ {-i,-} modProd m (modInv m (lua!!-(i,i)))
                            (modDif m (y!-i) 
                               (sumOf m (i+1,n) (\j -> modProd m (lua!!-(i,j)) (x!-j))))
           | i <- [n-1,n-2..one] ] 

    bds@(one,n) = vecBounds y
    bds'@((rl,cl),(rh,ch)) = matBounds lua
  in
    x

{-# SPECIALISE 
  homsolv :: Int -> SqMatrix Int -> Vector Int -> Vector Int
  #-}
homsolv :: (Integral a) => 
           a -> SqMatrix a -> Vector a -> Vector a
homsolv m a b = 
  let
    lua = lu m a
    (l,u) = split_lu m lua
    y = fwd_subst m l b
    x = bwd_subst m u y
  in 
  x

{-# SPECIALISE 
  det :: Int -> SqMatrix Int -> Int
  #-}
det :: (Integral a) => 
       a -> SqMatrix a -> a
det m a = let
          bds@((rl,cl),(rh,ch)) = matBounds a
          lua = lu m a
        in 
        foldl (\ z x -> modProd m z x) 1 [ lua!!-(j,j) | j <- [rl..rh] ]

mult_mat_vec :: (Integral a) => 
                a -> SqMatrix a -> Vector a -> Vector a
mult_mat_vec q m v = vector 
                     [ {-i,-} foldl (\ z x -> modSum q z x) 0 [ (m!!-(i,j))*(v!-j) | j <- [cl..ch] ]
                     | i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = matBounds m
        
mult_mat_mat :: (Integral a) => 
                a -> SqMatrix a -> SqMatrix a -> SqMatrix a
mult_mat_mat q m1 m2= lolSqMatrix bds
                   [  [ {-(i,j),-} 
                       foldl (\ z x -> modSum q z x) 0  [ modProd q (m1!!-(i,k)) (m2!!-(k,j)) | k <- [rl..rh] ]  
                     | j <- [cl..ch] ] 
                   | i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = matBounds m1

mult_vec_vec q v1 v2 = foldl (\ z x -> modSum q z x) 0 
                             [ modProd q (v1!-i) (v2!-i) | i<-[l..h] ]
                     where bds@(l,h) = vecBounds v1

split_lu :: (Integral a) => 
            a -> SqMatrix a -> (SqMatrix a, SqMatrix a)
split_lu _ lu = 
  let 
    l = lolSqMatrix bnds 
         [ [ {-(i,j),-} x | j <- [cl..ch] , 
                            let x = if i==j then 1 else (if i<j then 0 else lu!!-(i,j)) ]
         | i <- [rl..rh] ]
    u = lolSqMatrix bnds 
         [ [ {-(i,j),-} x | j <- [cl..ch] , 
                            let x = if i>j then 0 else lu!!-(i,j) ]
         | i <- [rl..rh] ]
    bnds@((rl,cl),(rh,ch)) = matBounds lu
  in
  (l,u)

check_lu :: (Integral a, NFData a) => 
            a -> SqMatrix a -> SqMatrix a -> Bool
check_lu q lu m = 
  let
    (l,u) = split_lu q m
  in
  if (mult_mat_mat q l u) == m
    then True
    else False

do_lu :: (Integral a, NFData a) => 
            a -> SqMatrix a -> IO ()
do_lu q m = do
            let x = lu q m
            let (l,u) = split_lu q x
            if (mult_mat_mat q l u)==m
              then putStrLn "OK"
              else putStrLn "WRONG"


            
-- ---------------------------------------------------------------------------
{-
-- a1 :: Array (Int, Int) Float
-- a1 = array ((1,1),(3,3)) 
--        [((1,1) ,  1.0), ((1,2) ,  2.0), ((1,3) ,  3.0), 
--         ((2,1) ,  4.0), ((2,2) , -3.0), ((2,3) ,  2.0), 
-- 	((3,1) ,  7.0), ((3,2) , -6.0), ((3,3) ,  0.0)]
-- -- det a1 = 31

a11 :: Array (Int, Int) (Ratio Integer)
a11 = array ((1,1),(3,3)) 
       [((1,1) ,  1 % 2), ((1,2) ,  2 % 3), ((1,3) ,  3 % 2), 
        ((2,1) ,  4 % 1), ((2,2) , -3 % 2), ((2,3) ,  2 % 5), 
	((3,1) ,  7 % 2), ((3,2) , -6 % 5), ((3,3) ,  0 % 1)]

listToListList :: Int -> [a] -> [[a]]
listToListList c m 
 | length m <= c = [m]
 | otherwise = c1 : listToListList c resto
  where (c1,resto) = splitAt c m

lolSqMatrix bds@((xlo,ylo), (xhi, yhi)) zss = 
  array bds (concat [ [ ((x,y), (zss!!(x-1))!!(y-1)) | y <- [ylo..yhi] ] | x <- [xlo..xhi] ])
            
lolArr bds@((xlo,ylo), (xhi, yhi)) zss = 
  array bds (concat [ [ ((x,y), (zss!!(x-1))!!(y-1)) | y <- [ylo..yhi] ] 
                    | x <- [xlo..xhi] ])
-}
