{-DA NOTE: Everything to do with difference arrays commented out!!! -}

{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}

-----------------------------------------------------------------------------
-- |
-- Module:    EvalStrategies
-- Author:    Patrick Maier <P.Maier@hw.ac.uk>
-- History:   Mar 19 2010: first floated
--            Apr 01 2010: major rewrite
--                         * proper parBuffer strategies
--                         * better clustering support
--
-- This module is an attempt to rewrite Simon Marlow's
-- Control.Parallel.Strategies (included in package parallel-2.2.0.1).
-- The name `EvalStrategies' is chosen not to clash with `Strategies'.
--
-- Main differences to Simon's Strategies:
-- 
-- * A leaner Eval monad (which does satisfy the monad laws).
--   Below, I'll give an informal rationale for the new strategy type.
--
-- * NFData is back. (Not sure why it was gone in Simon's Strategies.)
--
-- Caveat: This module is not yet tested thoroughly.
--         At the end there are a number of hand-reduced examples of the
--         generic `evalList' strategy. Provided I don't have a flawed
--         understanding of reductions in Haskell, these examples should
--         demonstrate that evalList behaves as expected.


-------------------------------------------------------------------------------
-- The Evolution of the Strategy Type
--
-- There are, as of now, 3 variants of the Strategy type
-- synonym. (Well, 3.5 if you want to be precise, but I'll discount
-- Simon's type because mine is very similar (and I can explain mine
-- but not his).
--
-- * There is the old definition
--   > type Strategy a = a -> ()
--   from the "Algorithm+Strategies=Parallelism" paper
--   <http://www.macs.hw.ac.uk/~dsg/gph/papers/html/Strategies/strategies.html>
--
-- * There is newer definition
--   > type Strategy a = a -> a
--   from the "Runtime Support for Multicore Haskell" paper
--   <http://www.haskell.org/~simonmar/papers/multicore-ghc.pdf>
--
-- * And there is the latest definition
--   > type Strategy a = a -> Eval a
--   > data Eval a = Done a
--   by myself (after a similar type used in Simon's Strategies module;
--   Simon's Eval type has two additional constructors, which are (in my
--   opinion) unnecessary and mess up his Eval monad).
--
-- I'll try to give a rationale for each of these types, based on my
-- informal understanding of parametricity afforded by parametric
-- polymorphism. (Caveat: This may be completely flawed.)
--
-- * Because polymorphism prevents any operations that actually inspect
--   values, the type "forall a . a -> ()" has only 3 (observably distinct)
--   functions: bottom (the everywhere undefined function), the constant
--   function and a strict version of the constant function. Of course,
--   if we were able to observe actual reductions, we would see many more
--   functions, evaluating their argument to varying degrees, sequentially
--   or in parallel. But the type makes it plain that the `side effect'
--   of evaluation is all these functions are about.
--
--   Actually, only two of these classes of functions provide reasonable
--   strategies: bottom is obviously useless. The other two, however,
--   are complimentary: The (non-strict) constant functions provide
--   strategies that `advise' evaluation (eg. in parallel, or no evaluation
--   at all) whereas the strict constant functions `mandate' evaluation
--   (eg. force weak head normal form, or even full normal form).
--
--   Were it not for the subtle issues with garbage collection identified in
--   "Runtime Support for Multicore Haskell", the type "forall a . a -> ()"
--   could arguably continue to be seen as *the* type of strategies.
--
-- * "Runtime Support for Multicore Haskell" proposed to saddle the
--   `side effect' of strategic evaluation on top of an identity function
--   rather than on a trivial constant function, hence the proposed type
--   "forall a . a -> a". Again, polymorphism tells us that this is
--   the type of identity functions (and of bottom, of course) --- a function
--   of this type simply can't do anything else with its argument other
--   than return it. However, that means that all functions of this *must*
--   be strict, hence all strategies *must* be strict.
--
--   It is the last point really that defeats this proposal. Non-strict
--   strategies (`advising' evaluation without forcing it) are too
--   important to be left out of a non-strict language.
--
-- * The newest proposal looks most complicated. The type of strategies is
--   "forall a . a -> Eval a" where "Eval a" is an algebraic datatype with
--   a single constructor "Done :: a -> Eval a". Because of this single
--   constructor (and again because of polymorphism), a function of type
--   "forall a . a -> Eval a" can't do anything other (on top of evaluation
--   as a `side effect') than wrapping "Done" around its argument and
--   returning the result. In spirit, this is still an identity function.
--   Except that it does not have to be strict.
--
--   With "forall a . a -> Eval a", we are back to our three classes of
--   (observably distinct) functions: bottom, Done-wrapped identity, and
--   strict Done-wrapped identity. The latter two correpond to the non-strict
--   resp. strict strategies afforded by the old strategies type
--   "forall a . a -> ()".
--
--   The Eval type constructor can also be made into a monad (rather trivially)
--   and hence into an applicative functor, thus providing the basis for
--   applying strategies during traversals over instances of the Traversable
--   class. (Thanks for that insight, Simon.)


-------------------------------------------------------------------------------

module EvalStrategies where
-- TODO: exports

import Prelude hiding (concat)
-- BEWARE: currently uses a local Parallel.hs -- HWL
import Parallel
-- import Control.Parallel
-- import ParallelTA (pseq, par)
import Control.Applicative
import Control.Monad
import Data.Traversable
import qualified Data.List (concat)
import Data.List (transpose)
import Data.Foldable (toList)
import Data.Sequence (Seq, (|>), fromList, ViewL(EmptyL,(:<)), viewl)
--DA import Data.Array.Diff (DiffArray, listArray, (!), (//)) 
-- BEWARE: currently uses a local Diff.hs (taken from ghc-6.10.1) -- HWL
import Diff (DiffArray, listArray, (!), (//)) 

infixr 9 `dot`     -- same as (.)
infixl 0 `using`   -- lowest precedence and associate to the left

-------------------------------------------------------------------------------
-- Eval monad

data Eval a = Done a

unEval :: Eval a -> a
unEval (Done x) = x

instance Functor Eval where
  fmap f x = x >>= return . f

instance Applicative Eval where
  pure x = return x
  (<*>) = ap

instance Monad Eval where
  return  = Done
  m >>= k = case m of
              Done x -> k x

-- The Eval monad satisfies the monad laws.
--
-- (1) Left identity:
--     return x >>= f ==> Done x >>= f ==> f x
--
-- (2) Right identity:
--     (i)  m >>= return =*> Done u >>= return
--                       ==> return u
--                       ==> Done u <*= m
--     (ii) m >>= return =*> undefined >>= return
--                       ==> undefined <*= m
--
-- (3) Associativity:
--     (i)  (m >>= f) >>= g =*> (Done u >>= f) >>= g
--                          ==> f u >>= g <== (\x -> f x >>= g) u
--                                        <== Done u >>= (\x -> f x >>= g)
--                                        <*= m >>= (\x -> f x >>= g)
--     (ii) (m >>= f) >>= g =*> (undefined >>= f) >>= g
--                          ==> undefined >>= g
--                          ==> undefined <== undefined >>= (\x -> f x >>= g)
--                                        <*= m >>= (\x -> f x >>= g)
--
-- Note: The Eval monad is essentially the Lift monad from Iavor Diatchki's
--       monadLib package
--       <http://hackage.haskell.org/packages/archive/monadLib/3.6.1/doc/html/src/MonadLib.html>


-------------------------------------------------------------------------------
-- Strategies

type Strategy a = a -> Eval a

-- | Evaluate a value using the given strategy
using :: a -> Strategy a -> a
using x strat = unEval (strat x)

-- | Compose two strategies sequentially; like function composition on strats
-- NB: * strat2 `dot` strat1 == strat2 . (`using` strat1)
--     * `dot' is not the same as the right-to-left Kleisli composition `(<=<)'
--       which happens to have the same type.
dot :: Strategy a -> Strategy a -> Strategy a
strat2 `dot` strat1 = strat2 . unEval . strat1


-------------------------------------------------------------------------------
-- Basic strategies

-- | r0 performs *no* evaluation
r0 :: Strategy a
r0 x = Done x

-- | rwhnf evaluates its argument to Weak Head Normal Form
rwhnf :: Strategy a
rwhnf x = x `pseq` Done x

-- | rpar sparks its argument (for evaluation in parallel)
rpar :: Strategy a
rpar x = x `par` Done x


-------------------------------------------------------------------------------
-- Normalform strategies

class NFData a where
    -- | rnf reduces its argument to (head) normal form.
    rnf :: Strategy a
    -- Default method. Useful for base types.
    -- A specific method is necessay for constructed types.
    rnf = rwhnf

instance NFData Int
instance NFData Integer
instance NFData Float
instance NFData Double
instance NFData Char
instance NFData Bool
instance NFData ()

instance (NFData a, NFData b) => NFData (a,b) where
    rnf = evalPair rnf rnf

instance (NFData a, NFData b, NFData c) => NFData (a,b,c) where
    rnf = evalTriple rnf rnf rnf

-- TODO: more tuples

instance (NFData a, NFData b, NFData c, NFData d, NFData e) => NFData (a,b,c,d,e) where
    rnf = evalQuintuple rnf rnf rnf rnf rnf

{- currently gives problems with overlapping instances.
   Currently using explicit List and Maybe instances instead 
instance (NFData a, Traversable t) => NFData (t a) where
    rnf = evalTraversable rnf
-}

-- These instances should be subsumed by the instances for Traversables.
instance NFData a => NFData (Maybe a) where
    rnf = evalMaybe rnf
instance NFData a => NFData [a] where
    rnf = evalList rnf


-------------------------------------------------------------------------------
-- Strategies for maybes

-- | Evaluate the contents of a maybe according to the given strategy
evalMaybe :: Strategy a -> Strategy (Maybe a)
evalMaybe strat  (Just x) = strat x >>= \x' -> return (Just x')
evalMaybe _strat Nothing  = return Nothing


-------------------------------------------------------------------------------
-- Strategies for tuples

-- | Evaluate the components of a pair according to given strategies
evalPair :: Strategy a -> Strategy b -> Strategy (a,b)
evalPair strat1 strat2 (x1,x2) = pure (,) <*> strat1 x1 <*> strat2 x2
-- This is the same as:
-- evalPair strat1 strat2 (x1,x2) = strat1 x1 >>= \x1' ->
--                                  strat2 x2 >>= \x2' ->
--                                  return (x1',x2')

-- | Evaluate the components of a triple according to given strategies
evalTriple :: Strategy a -> Strategy b -> Strategy c -> Strategy (a,b,c)
evalTriple strat1 strat2 strat3 (x1,x2,x3)
  = pure (,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3

-- | Evaluate the components of a quintuple according to given strategies
evalQuintuple :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy (a,b,c,d,e)
evalQuintuple strat1 strat2 strat3 strat4 strat5 (x1,x2,x3,x4,x5)
  = pure (,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5

-- Some more well-known tuple strategies:
-- (What's being said about the naming scheme for list strategies below
-- applies to tuple strategies as well.)

seqPair, parPair :: Strategy a -> Strategy b -> Strategy (a,b)
seqPair = evalPair
parPair strat1 strat2
  = evalPair (rpar `dot` strat1) (rpar `dot` strat2)

seqTriple, parTriple :: Strategy a -> Strategy b -> Strategy c -> Strategy (a,b,c)
seqTriple = evalTriple
parTriple strat1 strat2 strat3
  = evalTriple (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3)


-------------------------------------------------------------------------------
-- Strategies for traversable data types

-- | Evaluate each element of a traversable data structure according to
-- the given strategy
evalTraversable :: Traversable t => Strategy a -> Strategy (t a)
evalTraversable = traverse

-- Some more well-known strategies for Travesables:

seqTraverse :: Traversable t => Strategy a -> Strategy (t a)
seqTraverse = evalTraversable
parTraverse strat = evalTraversable (rpar `dot` strat)

-- a thresholding strategy would be cool: provided with a depth function
-- to calculate how deep in the structure we are, either create parallelism
-- or don't; e.g. create only parallelism in the top 2 levels of a tree -- HWL


-------------------------------------------------------------------------------
-- Strategies for lists

-- | Evaluate each element of a list according to the given strategy;
-- should be specialisation of evalTraversable
evalList :: Strategy a -> Strategy [a]
evalList strat []     = return []
evalList strat (x:xs) = strat x >>= \x' ->
                        evalList strat xs >>= \xs' ->
                        return (x':xs')

-- | Evaluate the first n elements of a list according to the given strategy
evalListN :: Int -> Strategy a -> Strategy [a]
evalListN 0   _strat xs     = return xs
evalListN !_n _strat []     = return []
evalListN !n  strat  (x:xs) = strat x >>= \x' ->
                              evalListN (n-1) strat xs >>= \xs' ->
                              return (x':xs')
-- This is the same as:
-- evalListN n strat = evalListTo n (evalList strat)

-- | Evaluate the prefix of length n of a list according to the given strategy
evalListTo :: Int -> Strategy [a] -> Strategy [a]
evalListTo n strat xs = let (ys,zs) = splitAt n xs in
                        strat ys >>= \ys' ->
                        return (ys' ++ zs)

-- | Skip over the first n elements and evaluate the remaining list
-- according to strat
evalListFrom :: Int -> Strategy [a] -> Strategy [a]
evalListFrom 0   strat  xs     = strat xs
evalListFrom !_n _strat []     = return []
evalListFrom !n  strat  (x:xs) = evalListFrom (n-1) strat xs >>= \xs' ->
                                 return (x:xs')

-- | Evaluate the nth element of a list (if there is such) according to
-- the given strategy
evalListNth :: Int -> Strategy a -> Strategy [a]
evalListNth n strat = evalListFrom n (evalListN 1 strat)
-- This is the same as:
-- evalListNth 0   strat  (x:xs) = strat x >>= \x' ->
--                                 return (x':xs)
-- evalListNth !_n _strat []     = return []
-- evalListNth !n  strat  (x:xs) = evalListNth (n-1) strat xs >>= \xs' ->
--                                 return (x:xs')

 
-------------------------------------------------------------------------------
-- Strategies for rolling buffers

{-
-- This is copied from Simon's Strategies module.
-- It compiles but has two defects:
-- * Return type of parBuffer is `[a] -> [a]' rather than `Strategy [a]'
-- * parBuffer always reduces buffer elements to WHNF even if the strategy
--   argument is `r0'.

parBuffer :: Int -> Strategy a -> [a] -> [a]
parBuffer n strat xs = map (`using` strat) xs `using` parBufferWHNF n

parBufferWHNF :: Int -> Strategy [a]
parBufferWHNF n0 xs0 = return (ret xs0 (start n0 xs0))
  where
    ret (x:xs) (y:ys) = y `par` (x : ret xs ys)
    ret xs     _      = xs

    start 0   ys     = ys
    start !_n []     = []
    start !n  (y:ys) = y `par` start (n-1) ys
-}


{- BROKEN: This code compiles and appears to work, but it still fails to
     record all sparks in the data structure. Might lead to loss of
     parallelism when GHC ceases to treat sparks as roots for GC.
-- | Rolling buffer of length n, evaluating the n-th element according
-- to the given strategy when the head is demanded.
-- This version should be more efficient (requires only a single traversal
-- of the list) but uses an unclean (ie. non-identity) strategy internally.
evalBuffer :: Int -> Strategy a -> Strategy [a]
evalBuffer n strat xs
  = return (map (fst . (`using` evalPair r0 (evalMaybe strat))) (zip xs ys))
      where ys = map Just (xs `using` dropN n) ++ repeat Nothing
            -- dropN :: Int -> Strategy [a]
            -- NB: dropN n is *not* an identity function; use only if it
            --     is certain that dropped elements do not become garbage
            dropN 0   zs     = return zs
            dropN !_n []     = return []
            dropN !n  (z:zs) = strat z >> dropN (n-1) zs

-- Some more well-known buffer strategies:
parBuffer :: Int -> Strategy a -> Strategy [a]
parBuffer n strat = evalBuffer n (rpar `dot` strat)
-}


-- NB: In a way, rolling buffer strategies are the equivalent to
--     list strategies, but for lazy or infinite lists. The type of a
--     rolling buffer strategy is `Strategy [a]', ie. a rolling buffer
--     presents itself as a (lazy) list. However, unlike lazy lists (which
--     trigger the evaluation of the head cell only when it is demanded),
--     a rolling buffer of size `n' triggers evaluation of the n-th cell
--     when the head is demanded. As a result, the buffer keeps up to n
--     list cells under evaluation.

-- The type signature of a generic rolling buffer strategy constructor is
-- > evalBuffer :: Int -> Strategy [a] -> Strategy a -> Strategy [a]
-- The strategy `evalBuffer n strat0 strat' creates a buffer of size `n',
-- where the list strategy `strat0' is applied to the prefix of the input list
-- used to fill the buffer initially, and the strategy `strat' is applied to
-- every later input list element before pushing it into the buffer.
-- NB: Rolling buffers are lazy, but initially filling them is not, ie. the
--     first element of a buffered list is not available before the buffer
--     is filled. This may lead to problems when using large buffers, or
--     when evaluating the spine of the input list is costly.

-- Currently, there are 3 implementations of rolling buffers, differing
-- in the data structures used to actually store the buffer.
--
-- * evalBufferL uses no special data structure; rather the buffering
--   functionality is implemented using the above list strategies.
--   Inefficient (at least for large buffer sizes): requires multiple
--   traversals of the input list.
--
-- * evalBufferQ uses a functional queue (with O(1) amortised update cost)
--   to store the buffer.
--
-- * evalBufferA uses a difference array (with O(1) update cost)
--   to store the buffer.


-- | Rolling buffer (using list strategies)
-- n:      buffer size
-- strat0: evaluation strategy for list filling buffer initially
-- strat:  evaluation strategy for elements pushed into buffer later
evalBufferL :: forall a. Int -> Strategy [a] -> Strategy a -> Strategy [a]
evalBufferL n strat0 strat xs
  = evalListTo n strat0 xs >>= \xs' ->
    return (roll xs')
      where roll :: [a] -> [a]
            roll []     = []
            roll (x:xs) = x:roll (xs `using` evalListNth n strat)

-- | Rolling buffer (using a functional queue)
-- n:      buffer size
-- strat0: evaluation strategy for list filling buffer initially
-- strat:  evaluation strategy for elements pushed into buffer later
-- NB: Elements that are under evaluation (or already evaluated) are kept in a
--     strict queue, the other elements in a lazy list. When the head of the
--     buffer is demanded, evaluate the head of the lazy list, stick it to the
--     right end of the queue, pop an element off the left end and return it.
--     The queue is realised by the Seq datatype, resulting in constant-time
--     operations (amortised, not worst case).
evalBufferQ :: forall a. Int -> Strategy [a] -> Strategy a -> Strategy [a]
evalBufferQ n strat0 strat xs
  = let (ys,zs) = splitAt n xs in
    strat0 ys >>= \ys' ->
    return (roll (fromList ys') zs)
      where roll :: Seq a -> [a] -> [a]
            roll q []     = toList q
            roll q (z:zs) = y':roll q' zs
                              where y':<q' = unEval (strat z >>= \z' ->
                                                     return (viewl (q|>z')))
                                    -- NB: Pattern match  can't fail
                                    -- because `q|>z' is non-empty.

-- | Rolling buffer (using a difference array)
-- n:      buffer size
-- strat0: evaluation strategy for list filling buffer initially
-- strat:  evaluation strategy for elements pushed into buffer later
-- requires: n > 0
-- NB: Elements that are under evaluation (or already evaluated) are kept in a
--     circular buffer, the other elements in a lazy list. When the head of
--     the buffer is demanded, return the oldest element in the circular buffer,
--     but only after evaluating the head of the lazy list, swapping it with
--     the oldest element, and advancing the buffer index.
--     The circular buffer is realised by a difference array, resulting in
--     constant-time operations (worst case).
evalBufferA :: forall a. Int -> Strategy [a] -> Strategy a -> Strategy [a]
evalBufferA n strat0 strat
  = return . roll . init
      where init :: [a] -> (Int, DiffArray Int (Maybe a), [Maybe a])
            init xs = (0, listArray (0,n-1) ys', zs')
                        where (ys,zs) = splitAt n xs
                              ys' = map Just (ys `using` strat0) ++ repeat Nothing
                              zs' = map Just zs ++ repeat Nothing

            roll :: (Int, DiffArray Int (Maybe a), [Maybe a]) -> [a]
            roll buf = case rotate buf of
                         (Nothing, _)    -> []
                         (Just x,  buf') -> x:roll buf'

            rotate :: (Int, DiffArray Int (Maybe a), [Maybe a]) -> (Maybe a, (Int, DiffArray Int (Maybe a), [Maybe a]))
            rotate (i, arr, feed) = (arr ! i, ((i+1) `mod` n, arr // [(i,z `using` evalMaybe strat)], zs))
                                      where z:zs = feed


-------------------------------------------------------------------------------
-- Clustering strategies

-- | Generic clustering strategy.
-- cluster: clustering function splitting input list into clusters
-- merge:   inverse of `cluster' (ie. merge . cluster == id)
-- strat:   strategy applied to the list of clusters
evalClusteredBy :: Cluster a -> Merge a -> Strategy [[a]] -> Strategy [a]
evalClusteredBy cluster merge strat xs = return (merge (cluster xs `using` strat))


-- | Clustering strategies for (strict) lists, applying `strat' to every cluster.
-- * evalListClusteredBy builds a strict, intermediate lists of clusters (which may lead
--   to non-constant space overhead)
-- * evalListClusteredBy' keeps the intermediate list of clusters lazy (which should
--   result in constant space overhead)
evalListClusteredBy, evalListClusteredBy' :: Cluster a -> Merge a -> Strategy [a] -> Strategy [a]
evalListClusteredBy cluster merge strat = evalClusteredBy cluster merge (evalList strat)
evalListClusteredBy' cluster merge strat = spineList . merge . map (`using` strat) . cluster


-- | Clustering strategies for rolling buffers.
-- n:      length of buffer (ie. number of clusters in buffer)
-- strat0: strategy applied to list of clusters filling buffer initially
-- strat:  strategy applied to clusters pushed into buffer later
evalBufferLClusteredBy, evalBufferQClusteredBy, evalBufferAClusteredBy
  :: forall a. Cluster a -> Merge a -> Int -> Strategy [[a]] -> Strategy [a] -> Strategy [a]
evalBufferLClusteredBy cluster merge n strat0 strat = evalClusteredBy cluster merge (evalBufferL n strat0 strat)
evalBufferQClusteredBy cluster merge n strat0 strat = evalClusteredBy cluster merge (evalBufferQ n strat0 strat)
evalBufferAClusteredBy cluster merge n strat0 strat = evalClusteredBy cluster merge (evalBufferA n strat0 strat)

-- Slightly optimized variant of evalBufferAClusteredBy (cutting out the Maybe)
-- requires: @cluster@ not to produce empty clusters, except at the end;
--           formally: @\xs -> all null $ dropWhile (not . null) $ cluster xs@
evalBufferAClusteredBy'
  :: forall a. Cluster a -> Merge a -> Int -> Strategy [[a]] -> Strategy [a] -> Strategy [a]
evalBufferAClusteredBy' cluster merge n strat0 strat
  = return . merge . roll . init . cluster
      where init :: [[a]] -> (Int, DiffArray Int [a], [[a]])
            init clusters = (0, listArray (0,n-1) (ys `using` strat0), zs)
                              where (ys,zs) = splitAt n (clusters ++ repeat [])

            roll :: (Int, DiffArray Int [a], [[a]]) -> [[a]]
            roll buf = case rotate buf of
                         ([], _)    -> []
                         (xs, buf') -> xs:roll buf'

            rotate :: (Int, DiffArray Int [a], [[a]]) -> ([a], (Int, DiffArray Int [a], [[a]]))
            rotate (i, arr, feed) = (arr ! i, ((i+1) `mod` n, arr // [(i,z `using` strat)], zs))
                                      where z:zs = feed

------------------------------------------------------------------------------
-- Clustering functions (based on our PPL paper)
-- http://www.macs.hw.ac.uk/~dsg/gph/papers/abstracts/hlpp01.html

{- 
These are the original definitions we used in that paper

class MMonadc  where
  munit  :: a -> c a
  mjoin  :: c (c a) -> c a
  mmap   :: (a -> b) -> (c a -> c b)

class (MMonad c) => Cluster c where
  cluster    :: Int -> c a -> c (c a)
  decluster  :: c (c a) -> c a
  lift       :: (c a -> b) -> (c (c a) -> c b)

  decluster = mjoin
  lift      = mmap

instance MMonad [] where
  munit x = [x]
  mjoin   = concat
  mmap    = map

instance Cluster [] where
  cluster   = splitAtN
  decluster = concat
  lift f    = map f

instance MMonad [] where
  munit x = [x]
  mjoin   = concat
  mmap    = map

instance Kluster [] where
  kluster   = chunk   -- used to be: splitAtN
  -- we get dekluster and lift for free
  -- dekluster = concat
  -- lift f    = map f


-- The old version built everything from an MMonad
-- To make use of the existing libs, we want to build from Monad
-- like this:

class Monad c => MMonad c  where
  munit  :: a -> c a
  mjoin  :: c (c a) -> c a
  mmap   :: (a -> b) -> (c a -> c b)

  munit = return
  mjoin = (>>= id)
  mmap  = liftM
  -- NB:   m >>= f = concatMap f m
  -- thus: join xs = concat xs = concatMap id xs = xs >>= id

-}

-- Here a slightly revised version, that should be easier to use

class (Monad c) => Kluster c where
  kluster    :: Int -> c a -> c (c a)
  dekluster  :: c (c a) -> c a
  lift       :: (c a -> b) -> (c (c a) -> c b)
  -- evalK      :: Strategy a -> Strategy (c a)

  dekluster = join
  lift      = liftM
  -- we require: dekluster . kluster n = id

{-
class  (MMonad m, MonadPlus m) => MMonadPlus m  where
  mzero_  :: m a
  mplus_  :: m a -> m a -> m a

class MMonadPlus m => MMonadFold m where
  mfold :: (m a -> m a -> m a) -> m a -> m (m a) -> m a
-}

instance Kluster [] where
  kluster = chunk
  -- evalK   = evalList

-- | Generic clustering strategy, using an overloaded cluster/decluster pair.
-- n:       cluster size
-- strat:   strategy applied to a cluster
evalKlustered :: Kluster c => Int -> Strategy (c (c a)) -> Strategy (c a)
evalKlustered n strat xs = return (dekluster (kluster n xs `using` strat))

-- the usual parListChunk, now implemented via klustering; just for testing KLuster class
parListChunkK :: Int -> Strategy a -> Strategy [a]
parListChunkK n strat = evalKlustered n (evalList (rpar `dot` seqList strat))

-- parChunkK generalises over the concrete data type
parChunkK :: (Kluster c, Traversable c) => Int -> Strategy a -> Strategy (c a)
-- parChunkK n strat = evalKlustered n ( {- liftStrat -} (rpar `dot` evalK {- liftStrat -} strat))
parChunkK n strat = evalKlustered n (evalTraversable (rpar `dot` evalTraversable strat))

-- parChunkK generalises over the concrete data type
{-
foldMapK :: (Kluster c, Traversable c, Foldable c) => Int -> Strategy a -> Strategy ((c a), a)
foldMapK f strat = \ (xs,x)  -> evalKlustered n (evalTraversable (rpar `dot` evalTraversable strat)) xs
                                evalFoldable f{-strat spec the order of combining results, eg. tree-over-list-} x
-}

-- tinkering with lifting strategies
liftStrat :: (Kluster c) => Strategy a -> Strategy (c a)
liftStrat s = \ xss -> return (liftM (unEval . s) xss)


{- Kluster instances for other data-types

data (Num a, NFData a) => Matrix a = M [[a]]

instance Monad Matrix where
  return x = M [[x]]
  m >>= f  = \ (M xss -> M (map (map f)))

instance Kluster Matrix where
  -- kluster :: (Num a, NFData a) => Int -> Matrix a -> Matrix (Matrix a)
  kluster n (M m1) = (M . (map wrap) . transpose . (map kluster) . 
                          transpose . map kluster) m1
    where wrap :: [[[a]]] -> Matrix a 
          wrap   = map (\ x -> M x)

  -- dekluster :: (Num a, NFData a) => Matrix (Matrix a) -> Matrix a
  dekluster (M m1) = (M . (map dekluster) . transpose . 
                          (map dekluster) . transpose . (map unwrap)) m1
    where unwrap :: Matrix a -> [[[a]]]
          unwrap = map (\ (M y) -> y )

  lift f (M m) = (M . map (map f)) m
-}


------------------------------------------------------------------------------
-- Clustering functions

-- | Type of clustering functions
type Cluster a = [a] -> [[a]]

-- | Type of merge functions (inverses of clustering functions)
type Merge a = [[a]] -> [a]

-- NB: The type of clustering functions could be generalised from
--     @forall a. [a] -> [[a]]@ to @forall a b. [a] -> b@, or to instances
--     thereof. This would provide greater flexibility for clustering
--     operations. (Not sure that's needed.)


-- | @'chunk' n xs@ chops @xs@ into chunks of length (up to) @n@, for @n > 0@.
-- For example,
-- 
-- > chunk 3 [1..8] == [[1,2,3],[4,5,6],[7,8]]
-- > chunk 8 [1..3] == [[1,2,3]]
chunk :: Int -> Cluster a
chunk _n [] = []
chunk n  xs = ys:chunk n zs where (ys,zs) = splitAt n xs


-- | @'concat'@ is an inverse of @'chunk' n@.
-- That is,
--
-- > (concat . chunk n) xs == xs
--
-- Moreover, @'chunk' n@ is an inverse of @'concat'@ for lists of lists
-- satisfying the first uniform-length property @ul1@, that is
--
-- > not (ul1 xss) || (chunk (length (head xss)) . concat) xss == xss
-- >   where ul1 []  = True
-- >         ul1 xss = all (== k) $ map length $ init xss
-- >                     where k = length (head xss)
--
-- NB: @'chunk' n xs@ satisfies property @ul1@, for all @n@ and @xs@.
concat :: Merge a
concat = Data.List.concat


-- | @'slice' n xs@ cuts @xs@ into (up to) @n@ slices, for @n > 0@.
-- For example,
--
-- > slice 3 [1..8] == [[1,4,7],[2,5,8],[3,6]]
-- > slice 8 [1..3] == [[1],[2],[3]]
slice :: Int -> Cluster a
slice n = transpose . chunk n

-- | Variant of @'slice'@; used to be called @unshuffle@.
-- NB: @slice' n xs@ always yields @n@ slices but trailing slices may be empty.
-- For example,
--
-- > slice' 3 [1..8] == [[1,4,7],[2,5,8],[3,6]]
-- > slice' 8 [1..3] == [[1],[2],[3],[],[],[],[],[]]
slice' :: Int -> Cluster a
slice' n xs = [takeEachN (drop i xs) | i <- [0..n-1]]
                where takeEachN []     = []
                      takeEachN (x:xs) = x : takeEachN (drop (n-1) xs)


-- | @'shuffle'@ is an inverse of @'slice' n@.
-- Proof:
--
-- > (shuffle . slice n) xs
-- >   == (concat . transpose . transpose . chunk n) xs
-- >   == (concat . chunk n) xs
-- >   == xs
--
-- Moreover, @'slice' n@ is an inverse of @'shuffle'@ for lists of lists
-- satisfying the second uniform-length property @ul2@, that is
--
-- > not (ul2 xss) || (slice (length xss) . shuffle) xss == xss
-- >   where ul2 []  = True
-- >         ul2 xss = all (== (k-1)) $ dropWhile (== k) $ map length xss
-- >                     where k = length (head xss)
--
-- NB: The list @'slice' n xs@ satisfies property @ul2@, for all @n@ and @xs@.
--     The function @map (chunk m)@ preserves property @ul2@, for all @m@.
shuffle :: Merge a
shuffle = concat . transpose

-- | Variant of @'shuffle'@.
shuffle' :: Merge a
shuffle' xss = case concat (map (take 1) xss) of
                 []    -> []
                 heads -> heads ++ shuffle' (map (drop 1) xss)

-- | Another variant of @'shuffle'@ (apparently coming from to Eden).
shuffle'' :: Merge a
shuffle'' xxs | and (map null xxs) = []
	      | otherwise  = mymaphead xxs ++ shuffle'' (map mytail xxs)
                               where mymaphead []           = []
		                     mymaphead ([]:xxs)     = mymaphead xxs
		                     mymaphead ((x:xs):xxs) = x : mymaphead xxs
                                     mytail [] = []
                                     mytail xs = tail xs


-- | @'sliceChunk' n m xs@ cuts @xs@ into @n@ slices, chops each slice into
-- chunks of size @m@, and shuffles the chunked slices, for @n > 0@ and @m > 0@.
-- That's a conjugation, wow!
-- For example,
--
-- > sliceChunk 3 3 [1..8] == [[1,4,7],[2,5,8],[3,6]] == slice 3 [1..8]
-- > sliceChunk 3 2 [1..8] == [[1,4],[2,5],[3,6],[7],[8]]
-- > sliceChunk 3 1 [1..8] == [[1],[2],[3],[4],[5],[6],[7],[8]] == chunk 1 [1..8]
sliceChunk :: Int -> Int -> Cluster a
sliceChunk n m = shuffle . map (chunk m) . slice n


-- | @'shuffleChunk' n@ is the left inverse of @'sliceChunk' n m@.
-- Proof:
--
-- > (shuffleChunk n . sliceChunk n m) xs
-- >   == (shuffle . map concat . slice n . shuffle . map (chunk m) . slice n) xs
-- >   == (shuffle . map concat . map (chunk m) . slice n) xs
-- >   == (shuffle . map (concat . chunk m) . slice n) xs
-- >   == (shuffle . map id . slice n) xs
-- >   == (shuffle . slice n) xs
-- >   == xs
--
-- Note that the second equality holds because @slice n@ is an inverse of
-- @shuffle@ (since @map (chunk m) $ slice n xs@ satisfies property @ul2@).
shuffleChunk :: Int -> Merge a
shuffleChunk n = shuffle . map concat . slice n

-- | Variant of @'shuffleChunk'@.
shuffleChunk' :: Int -> Merge a
shuffleChunk' n xss = case splitAt n xss of
                        ([],  _)   -> []
                        (yss, zss) -> shuffle yss ++ shuffleChunk' n zss


------------------------------------------------------------------------------
-- More well-known list and buffer strategies; names (and in most cases type
-- signatures) derived from the old strategies module

spineList :: Strategy [a]
spineList = evalList r0

seqList, parList :: Strategy a -> Strategy [a]
seqList = evalList
parList strat = evalList (rpar `dot` strat)

seqListN, parListN, parListNth :: Int -> Strategy a -> Strategy [a]
seqListN = evalListN
parListN   n strat = evalListN   n (rpar `dot` strat)
parListNth n strat = evalListNth n (rpar `dot` strat)

parMap :: Strategy b -> (a -> b) -> [a] -> [b]
parMap strat f xs = map f xs `using` parList strat

-- parList operating on chunks of input list (to increase granularity)
parListChunk, parListChunk' :: Int -> Strategy a -> Strategy [a]
parListChunk  n strat = evalListClusteredBy  (chunk n) concat (rpar `dot` seqList strat)
parListChunk' n strat = evalListClusteredBy' (chunk n) concat (rpar `dot` seqList strat)

-- parList operating on slices of input list (to increase granularity)
parListSlice, parListSlice' :: Int -> Strategy a -> Strategy [a]
parListSlice  n strat = evalListClusteredBy  (slice n) shuffle (rpar `dot` seqList strat)
parListSlice' n strat = evalListClusteredBy' (slice n) shuffle (rpar `dot` seqList strat)

parBufferL, parBufferQ, parBufferA :: Int -> Strategy a -> Strategy [a]
parBufferL n strat = evalBufferL n (parList strat) (rpar `dot` strat)
parBufferQ n strat = evalBufferQ n (parList strat) (rpar `dot` strat)
parBufferA n strat = evalBufferA n (parList strat) (rpar `dot` strat)

-- parBuffer operating on chunks of input list (to inc granularity and throughput)
-- n:     buffer size (ie. number of chunks in buffer)
-- m:     chunk size
-- strat: strategy applied to chunk before pushing it into buffer
parBufferLChunk, parBufferQChunk, parBufferAChunk, parBufferAChunk' :: Int -> Int -> Strategy [a] -> Strategy [a]
parBufferLChunk  n m strat = evalBufferLClusteredBy  (chunk m) concat n (parList strat) (rpar `dot` strat)
parBufferQChunk  n m strat = evalBufferQClusteredBy  (chunk m) concat n (parList strat) (rpar `dot` strat)
parBufferAChunk  n m strat = evalBufferAClusteredBy  (chunk m) concat n (parList strat) (rpar `dot` strat)
parBufferAChunk' n m strat = evalBufferAClusteredBy' (chunk m) concat n (parList strat) (rpar `dot` strat)

-- parBuffer operating on chunks of slices of input list (to increase granularity and
-- buffer throughput without compromising load balancing)
-- n:     buffer size (ie. number of chunks in buffer)
-- m:     number of slices (and stride between elements)
-- k:     chunk size
-- strat: strategy applied to chunk before pushing it into buffer
parBufferLSliceChunk, parBufferQSliceChunk, parBufferASliceChunk, parBufferASliceChunk' 
  :: Int -> Int -> Int -> Strategy [a] -> Strategy [a]
parBufferLSliceChunk  n m k strat
  = evalBufferLClusteredBy  (sliceChunk m k) (shuffleChunk m) n (parList strat) (rpar `dot` strat)
parBufferQSliceChunk  n m k strat
  = evalBufferQClusteredBy  (sliceChunk m k) (shuffleChunk m) n (parList strat) (rpar `dot` strat)
parBufferASliceChunk  n m k strat
  = evalBufferAClusteredBy  (sliceChunk m k) (shuffleChunk m) n (parList strat) (rpar `dot` strat)
parBufferASliceChunk' n m k strat
  = evalBufferAClusteredBy' (sliceChunk m k) (shuffleChunk m) n (parList strat) (rpar `dot` strat)


-- The above list strategies are not optimally compositional --- note eg.
-- the type of parListChunk, expecting a strategy on elements rather
-- than on chunks, and compare to parBufferChunk --- and awkwardly named
-- (a fact that will become obvious as soon as one tries to generalize these
-- strategies to the new `par' operators bpar, parAt, etc. Therefore,
-- I recommend deprecating these old strategies. (A convincing new naming
-- scheme has yet to be worked out.)


------------------------------------------------------------------------------
-- Strategic function application; copied from Simon's Strategies module
-- (going back to the Ur-Strategies module by Trinder, Loidl, Hammond et al.)

{-
These are very handy when writing pipeline parallelism asa sequence of
@$@, @$|@ and @$||@'s. There is no need of naming intermediate values
in this case. The separation of algorithm from strategy is achieved by
allowing strategies only as second arguments to @$|@ and @$||@.
-}

-- | Sequential function application. The argument is evaluated using
--   the given strategy before it is given to the function.
($|) :: (a -> b) -> Strategy a -> a -> b
f $| s  = \ x -> let z = x `using` s in z `pseq` f z

-- | Parallel function application. The argument is evaluated using
-- the given strategy, in parallel with the function application.
($||) :: (a -> b) -> Strategy a -> a -> b
f $|| s = \ x -> let z = x `using` s in z `par` f z

-- | Sequential function composition. The result of
-- the second function is evaluated using the given strategy, 
-- and then given to the first function.
(.|) :: (b -> c) -> Strategy b -> (a -> b) -> (a -> c)
(.|) f s g = \ x -> let z = g x `using` s in 
                    z `pseq` f z

-- | Parallel function composition. The result of the second
-- function is evaluated using the given strategy,
-- in parallel with the application of the first function.
(.||) :: (b -> c) -> Strategy b -> (a -> b) -> (a -> c)
(.||) f s g = \ x -> let z = g x `using` s in 
                    z `par` f z

-- | Sequential inverse function composition, 
-- for those who read their programs from left to right.
-- The result of the first function is evaluated using the 
-- given strategy, and then given to the second function.
(-|) :: (a -> b) -> Strategy b -> (b -> c) -> (a -> c)
(-|) f s g = \ x -> let z = f x `using` s in 
                    z `pseq` g z

-- | Parallel inverse function composition,
-- for those who read their programs from left to right.
-- The result of the first function is evaluated using the 
-- given strategy, in parallel with the application of the 
-- second function.
(-||) :: (a -> b) -> Strategy b -> (b -> c) -> (a -> c)
(-||) f s g = \ x -> let z = f x `using` s in 
                    z `par` g z


-------------------------------------------------------------------------------
-- Deprecated functions:
-- These may not record sparks (hence cause loss of parallelism in future)

demanding :: a -> b -> a
demanding = flip pseq

sparking :: a -> b -> a
sparking  = flip par


-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
{-
-- HAND-REDUCED EXAMPLES
--
-- One-step reduction (lazy, ie. at outermost redex) is denoted by ==>,
-- n-step reduction by =n>, and arbitrarily many (including 0) steps by =*>.
--
-- The reduction rule for `par' is
-- > e `par` e' ==> e'[e*/e]
-- where e'[e*/e] means "replace every occurence of e in e' with e*".
-- The * symbolises that the expression has been sparked.
-- Note: This simple minded semantics doesn't properly count the number
--       of sparks, it just tracks the live sparked expressions.

evalList strat []
==> return []
==> Done []

[] `using` evalList strat
==> unEval (evalList strat [])
=2> unEval (Done [])
==> []


evalList r0 (take 1 (repeat undefined))
=2> evalList r0 (undefined:take 0 (repeat undefined))
==> r0 undefined >>= \x' ->
    evalList r0 (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
==> Done undefined >>= \x' ->
    evalList r0 (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
=2> evalList r0 (take 0 (repeat undefined)) >>= \xs' ->
    return (undefined:xs')
==> evalList r0 [] >>= \xs' ->
    return (undefined:xs')
=2> Done [] >>= \xs' ->
    return (undefined:xs')
=2> return [undefined]
==> Done [undefined]


evalList rwhnf (take 1 (repeat undefined))
=2> evalList rwhnf (undefined:take 0 (repeat undefined))
==> rwhnf undefined >>= \x' ->
    evalList rwhnf (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
==> (undefined `pseq` Done undefined) >>= \x' ->
    evalList rwhnf (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
==> undefined


evalList rpar (take 1 (repeat undefined))
=2> evalList rpar (undefined:take 0 (repeat undefined))
==> rpar undefined >>= \x' ->
    evalList rpar (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
==> (undefined `par` Done undefined) >>= \x' ->
    evalList rpar (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
==> Done undefined* >>= \x' ->
    evalList rpar (take 0 (repeat undefined)) >>= \xs' ->
    return (x':xs')
=2> evalList rpar (take 0 (repeat undefined)) >>= \xs' ->
    return (undefined*:xs')
==> evalList rpar [] >>= \xs' ->
    return (undefined*:xs')
=2> Done [] >>= \xs' ->
    return (undefined*:xs')
=2> return [undefined*]
==> Done [undefined*]


evalList r0 [1+1,2+2]
==> r0 (1+1) >>= \x' ->
    evalList r0 [2+2] >>= \xs' ->
    return (x':xs')
==> Done (1+1) >>= \x' ->
    evalList r0 [2+2] >>= \xs' ->
    return (x':xs')
=2> evalList r0 [2+2] >>= \xs' ->
    return ((1+1):xs')
==> (r0 (2+2) >>= \x' ->
     evalList r0 [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return ((1+1):xs')
==> (Done (2+2) >>= \x' ->
     evalList r0 [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return ((1+1):xs')
=2> (evalList r0 [] >>= \xs' ->
     return ((2+2):xs') >>= \xs' ->
    return ((1+1):xs')
=2> (Done [] >>= \xs' ->
     return ((2+2):xs') >>= \xs' ->
    return ((1+1):xs')
=2> return [2+2] >>= \xs' ->
    return ((1+1):xs')
==> Done [2+2] >>= \xs' ->
    return ((1+1):xs')
=2> return [1+1,2+2]
==> Done [1+1,2+2]

[1+1,2+2] `using` evalList r0
==> unEval (evalList r0 [1+1,2+2])
=*> unEval (Done [1+1,2+2])
==> [1+1,2+2]


evalList rwhnf [1+1,2+2]
==> rwhnf (1+1) >>= \x' ->
    evalList rwhnf [2+2] >>= \xs' ->
    return (x':xs')
==> ((1+1) `pseq` Done (1+1)) >>= \x' ->
    evalList rwhnf [2+2] >>= \xs' ->
    return (x':xs')
==> (2 `pseq` Done 2) >>= \x' ->
    evalList rwhnf [2+2] >>= \xs' ->
    return (x':xs')
==> Done 2 >>= \x' ->
    evalList rwhnf [2+2] >>= \xs' ->
    return (x':xs')
=2> evalList rwhnf [2+2] >>= \xs' ->
    return (2:xs')
==> (rwhnf (2+2) >>= \x' ->
     evalList rwhnf [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return (2:xs')
==> (((2+2) `pseq` Done (2+2)) >>= \x' ->
     evalList rwhnf [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return (2:xs')
==> ((4 `pseq` Done 4) >>= \x' ->
     evalList rwhnf [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return (2:xs')
==> (Done 4 >>= \x' ->
     evalList rwhnf [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return (2:xs')
=2> (evalList rwhnf [] >>= \xs' ->
     return (4:xs') >>= \xs' ->
    return (2:xs')
=2> (Done [] >>= \xs' ->
     return (4:xs') >>= \xs' ->
    return (2:xs')
=2> return [4] >>= \xs' ->
    return (2:xs')
==> Done [4] >>= \xs' ->
    return (2:xs')
=2> return [2,4]
==> Done [2,4]

[1+1,2+2] `using` evalList rwhnf
==> unEval (evalList rwhnf [1+1,2+2])
=*> unEval (Done [2,4])
==> [2,4]


evalList rpar [1+1,2+2]
==> rpar (1+1) >>= \x' ->
    evalList rpar [2+2] >>= \xs' ->
    return (x':xs')
==> ((1+1) `par` Done (1+1)) >>= \x' ->
    evalList rpar [2+2] >>= \xs' ->
    return (x':xs')
==> Done (1+1)* >>= \x' ->
    evalList rpar [2+2] >>= \xs' ->
    return (x':xs')
=2> evalList rpar [2+2] >>= \xs' ->
    return ((1+1)*:xs')
==> (rpar (2+2) >>= \x' ->
     evalList rpar [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return ((1+1)*:xs')
==> (((2+2) `par` Done (2+2)) >>= \x' ->
     evalList rpar [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return ((1+1)*:xs')
==> (Done (2+2)* >>= \x' ->
     evalList rpar [] >>= \xs' ->
     return (x':xs') >>= \xs' ->
    return ((1+1)*:xs')
=2> (evalList rpar [] >>= \xs' ->
     return ((2+2)*:xs') >>= \xs' ->
    return ((1+1)*:xs')
=2> (Done [] >>= \xs' ->
     return ((2+2)*:xs') >>= \xs' ->
    return ((1+1)*:xs')
=2> return [(2+2)*] >>= \xs' ->
    return ((1+1)*:xs')
==> Done [(2+2)*] >>= \xs' ->
    return ((1+1)*:xs')
=2> return [(1+1)*,(2+2)*]
==> Done [(1+1)*,(2+2)*]

[1+1,2+2] `using` evalList rpar
==> unEval (evalList rpar [1+1,2+2])
=*> unEval (Done [(1+1)*,(2+2)*])
==> [(1+1)*,(2+2)*]


evalList (evalList rpar) [[1+1,2+2],[]]
==> evalList rpar [1+1,2+2] >>= \x' ->
    evalList (evalList rpar) [[]] >>= \xs' ->
    return (x':xs')
=*> Done [(1+1)*,(2+2)*] >>= \x' ->
    evalList (evalList rpar) [[]] >>= \xs' ->
    return (x':xs')
=2> evalList (evalList rpar) [[]] >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
==> (evalList rpar [] >>= \x' ->
     evalList (evalList rpar) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
=2> (Done [] >>= \x' ->
     evalList (evalList rpar) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
=2> (evalList (evalList rpar) [] >>= \xs' ->
     return ([]:xs')) >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
=2> (Done [] >>= \xs' ->
     return ([]:xs')) >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
=2> return [[]] >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
==> Done [[]] >>= \xs' ->
    return ([(1+1)*,(2+2)*]:xs')
=2> return [[(1+1)*,(2+2)*],[]]
==> Done [[(1+1)*,(2+2)*],[]]


(evalList rpar . map (`using` evalList rpar)) [[1+1,2+2],[]]
==> evalList rpar (map (`using` evalList rpar) [[1+1,2+2],[]])
==> evalList rpar (([1+1,2+2] `using` evalList rpar):map (`using` evalList rpar) [[]])
==> rpar ([1+1,2+2] `using` evalList rpar) >>= \x' ->
    evalList rpar (map (`using` evalList rpar) [[]]) >>= \xs' ->
    return (x':xs')
==> (([1+1,2+2] `using` evalList rpar) `par` Done ([1+1,2+2] `using` evalList rpar)) >>= \x' ->
    evalList rpar (map (`using` evalList rpar) [[]]) >>= \xs' ->
    return (x':xs')
==> Done ([1+1,2+2] `using` evalList rpar)* >>= \x' ->
    evalList rpar (map (`using` evalList rpar) [[]]) >>= \xs' ->
    return (x':xs')
=2> evalList rpar (map (`using` evalList rpar) [[]]) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> evalList rpar (([] `using` evalList rpar):map (`using` evalList rpar) []) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> (rpar ([] `using` evalList rpar) >>= \x' ->
     evalList rpar (map (`using` evalList rpar) []) >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> ((([] `using` evalList rpar) `par` Done ([] `using` evalList rpar)) >>= \x' ->
     evalList rpar (map (`using` evalList rpar) []) >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> (Done ([] `using` evalList rpar)* >>= \x' ->
     evalList rpar (map (`using` evalList rpar) []) >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> (evalList rpar (map (`using` evalList rpar) []) >>= \xs' ->
     return (([] `using` evalList rpar)*:xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> (evalList rpar [] >>= \xs' ->
     return (([] `using` evalList rpar)*:xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> (Done [] >>= \xs' ->
     return (([] `using` evalList rpar)*:xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> return [([] `using` evalList rpar)*] >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> Done [([] `using` evalList rpar)*] >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> return [([1+1,2+2] `using` evalList rpar)*,([] `using` evalList rpar)*]
==> Done [([1+1,2+2] `using` evalList rpar)*,([] `using` evalList rpar)*]


evalList (rpar . (`using` evalList rpar)) [[1+1,2+2],[]]
==> (rpar . (`using` evalList rpar)) [1+1,2+2] >>= \x' ->
    evalList (rpar . (`using` evalList rpar)) [[]] >>= \xs' ->
    return (x':xs')
==> rpar ([1+1,2+2] `using` evalList rpar) >>= \x' ->
    evalList (rpar . (`using` evalList rpar)) [[]] >>= \xs' ->
    return (x':xs')
==> (([1+1,2+2] `using` evalList rpar) `par` Done ([1+1,2+2] `using` evalList rpar)) >>= \x' ->
    evalList (rpar . (`using` evalList rpar)) [[]] >>= \xs' ->
    return (x':xs')
==> Done ([1+1,2+2] `using` evalList rpar)* >>= \x' ->
    evalList (rpar . (`using` evalList rpar)) [[]] >>= \xs' ->
    return (x':xs')
=2> evalList (rpar . (`using` evalList rpar)) [[]] >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> ((rpar . (`using` evalList rpar)) [] >>= \x' ->
     evalList (rpar . (`using` evalList rpar)) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> (rpar ([] `using` evalList rpar) >>= \x' ->
     evalList (rpar . (`using` evalList rpar)) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> ((([] `using` evalList rpar) `par` Done ([] `using` evalList rpar)) >>= \x' ->
     evalList (rpar . (`using` evalList rpar)) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> (Done ([] `using` evalList rpar)* >>= \x' ->
     evalList (rpar . (`using` evalList rpar)) [] >>= \xs' ->
     return (x':xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> (evalList (rpar . (`using` evalList rpar)) [] >>= \xs' ->
     return (([] `using` evalList rpar)*:xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> (Done [] >>= xs' ->
     return (([] `using` evalList rpar)*:xs')) >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> return [([] `using` evalList rpar)*] >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
==> Done [([] `using` evalList rpar)*] >>= \xs' ->
    return (([1+1,2+2] `using` evalList rpar)*:xs')
=2> return [([1+1,2+2] `using` evalList rpar)*,([] `using` evalList rpar)*]
==> Done [([1+1,2+2] `using` evalList rpar)*,([] `using` evalList rpar)*]
-}

------------------------------------------------------------------------------------------------------------------------
--                                                                                                 1         1         1
--       1         2         3         4         5         6         7         8         9         0         1         2
-- 456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
