{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE EmptyDataDecls #-}

-----------------------------------------------------------------------------
-- |
-- Module:    Control.Parallel.Clustering
-- Copyright   :  (c) The University of Glasgow 2001-2009
-- License     :  BSD-style (see the file libraries/base/LICENSE)
--
-- Maintainer  :  libraries@haskell.org
-- Stability   :  experimental
-- Portability :  portable
--
-- Clustering strategies and functions.
-- TODO: Better description.
--

module Control.Parallel.Clustering
       ( -- * Cluster class
         Cluster(..) 
         -- * strategy wrapper for explicating cluster type in evalCluster
       , ClusterStrategy(ClustStrat)
       -- clustering using instances of Cluster class
       -- using GADTs
       , evalCluster
       , parCluster
       -- withour GADTs
       , ClusterWith(C)
       , evalCluster0
       , parCluster0
       , parMapCluster0
       -- withour GADTs (final version in the Haskell10 paper)
       , ClusterWithX
       , evalClusterX
       , parClusterX
       , parMapClusterX
       , __

         -- * Clustering strategies
       , evalClusterBy   -- very unsafe; should prob not be exported

       -- TODO: add code for these again
       -- , evalClusterComb
       -- , parClusterComb

       -- Variants of clustering; only tmp
       , Cluster1(..), {- Cluster2(..), -} Cluster3(..), Cluster4(..), Cluster5(..)
       , evalCluster1, {- evalCluster2, -} evalCluster3, evalCluster4, evalCluster5
   
       -- ToDo: renable
       -- , evalTraversableCluster   -- obsolete with new Cluster class
       -- , parTraversableCluster    -- obsolete with new Cluster class
         
         -- * Clustering strategies for lists
       , evalListChunk  
       , parListChunk
       
       , evalListSlice
       , parListSlice
         
         -- * Clustering strategies for lazy lists
       , evalBufferChunk  
       , parBufferChunk
         
       , evalBufferChunk'
       , parBufferChunk'
         
       , evalBufferSliceChunk
       , parBufferSliceChunk
         
       , evalBufferSliceChunk'
       , parBufferSliceChunk'
         
         -- * Clustering functions for lists
       , chunk
    -- , concat   -- exported by the Prelude
       
       , slice
       , shuffle
       
       , sliceChunk
       , shuffleChunk
       ) where

import Prelude
import Control.Monad (liftM, join)
import Data.List (transpose)
import Data.Traversable (Traversable)
import Data.Foldable (fold)
import Data.Monoid
import Control.Parallel.Strategies

import qualified Data.Set -- only for testing instances of Cluster

-- --------------------------------------------------------------------------
-- Cluster class

class Monad c => Cluster0 c where
  cluster0    :: Int -> c a -> c (c a)
  decluster0  :: c (c a) -> c a
  lift0       :: (c a -> b) -> (c (c a) -> c b)

  decluster0 = join
  lift0      = liftM
  -- we require: decluster . cluster n == id


-- instance for lists with chunking
instance Cluster0 [] where
  cluster0 = chunk


-- apply a strategy to all components of a cluster
-- this should go into Strategies.hs
liftStrat :: (Cluster0 c) => Strategy a -> Strategy (c a)
liftStrat s = \ xss -> return (liftM (runEval . s) xss)

-- Various variants on clustering
--
-- V1: Traversable instead of Monad
--
class Traversable c => Cluster1 c where
  cluster1    :: Int -> c a -> c (c a)
  decluster1  :: c (c a) -> c a
  lift1       :: (c a -> b) -> (c (c a) -> c b)

  lift1 = fmap   -- c is a Functor
  -- we require: decluster1 . cluster1 n == id

-- instance for lists with chunking
instance Cluster1 [] where
  cluster1 = chunk
  decluster1 = concat
  lift1 = fmap
  
evalCluster1 :: Cluster1 c => Int -> Strategy (c a) -> Strategy (c a)
evalCluster1 n strat = evalClusterBy (cluster1 n) decluster1 (evalTraversable strat)

parCluster1 :: Cluster1 c => Int -> Strategy (c a) -> Strategy (c a)
parCluster1 n strat = evalCluster1 n (rpar `dot` strat)

--
-- V2: list of clusters (no context required for c)
--
{-
class Traversable a => Cluster2 c a where
  cluster2    :: Int -> a -> [a]
  decluster2  :: [a] -> a
  lift2       :: (a -> b) -> [a] -> [b]

  lift2 = fmap
  decluster2 = mconcat
  -- we require: decluster2 . cluster2 n == id
  
-- instance for lists with chunking
instance Cluster2 [] where
  cluster2 = chunk
  -- decluster2 = concat

evalCluster2 :: Cluster2 c => Int -> Strategy (c a) -> Strategy (c a)
evalCluster2 n strat = evalClusterBy (cluster2 n) decluster2 (evalList strat)

parCluster2 :: Cluster2 c => Int -> Strategy (c a) -> Strategy (c a)
parCluster2 n strat = evalCluster2 n (rpar `dot` strat)
-}

--
-- V3: Traversable of clusters (on context required for c)
-- exts reqd: MultiParamTypeClasses, ScopedTypeVariables
--
class Traversable d => Cluster3 c d where
  cluster3    :: Int -> c a -> d (c a)
  decluster3  :: d (c a) -> c a
  lift3       :: (c a -> b) -> d (c a) -> d b

  lift3 = fmap   -- d is a Functor
  -- we require: decluster3 . cluster3 n == id

-- instance for lists with chunking
instance Cluster3 [] [] where
  cluster3 = chunk
  decluster3 = concat

evalCluster3 :: forall a c d . Cluster3 c d => Int -> Strategy (c a) -> Strategy (c a)
evalCluster3 n strat = evalClusterBy (cluster3 n) decluster3 stratC
  where stratC = evalTraversable strat :: Strategy (d (c a))

-- parCluster3 :: forall a c d . Cluster3 c d => Int -> Strategy (c a) -> Strategy (c a)
parCluster3 n strat = evalCluster3 n (rpar `dot` strat)

--
-- V4: Traversable of clusters (clusters of arbitrary type)
-- exts reqd: MultiParamTypeClasses, ScopedTypeVariables
--
class Traversable c => Cluster4 a c {- | a -> c -}  where
  cluster4    :: Int -> a -> c a
  decluster4  :: c a -> a
  lift4       :: (a -> b) -> c a -> c b

  lift4 = fmap   -- c is a Functor
  -- we require: decluster4 . cluster4 n == id

-- instance for lists with chunking
instance Cluster4 [a] [] where
  cluster4 = chunk
  decluster4 = concat
  
evalCluster4 :: forall a c . Cluster4 a c => Int -> Strategy a -> Strategy a
evalCluster4 n strat = evalClusterBy (cluster4 n) decluster4 stratC
  where stratC = evalTraversable strat :: Strategy (c a)

-- parCluster4 :: Cluster4 a c => Int -> Strategy a -> Strategy a
-- parCluster4 n strat = evalCluster4 n (rpar `dot` strat)

-- with awkward functional dependency

class (Traversable c, Monoid a) => 
      Cluster5 a c | a -> c   where
  cluster5    :: Int -> a -> c a
  decluster5  :: c a -> a
  lift5       :: (a -> b) -> c a -> c b

  lift5 = fmap       -- c is a Functor
  decluster5 = fold  -- c is Foldable
  -- we require: decluster . cluster n == id

-- instance for lists with chunking
instance Cluster5 [a] [] where
  cluster5 = chunk

-- another instance, just for testing  
{-
instance Cluster [a] Data.Set.Set where
  cluster = Data.Set.fromList . chunk
-}
  
evalCluster5 :: forall a c. Cluster5 a c => Int -> Strategy a -> Strategy a
evalCluster5 n strat = evalClusterBy (cluster5 n) decluster5 stratC
  where stratC = evalTraversable strat :: Strategy (c a)

parCluster5 :: forall a c. Cluster5 a c => Int -> Strategy a -> Strategy a
parCluster5 n strat = -- evalCluster5 n (rpar `dot` strat)
                     evalClusterBy (cluster5 n) decluster5 stratC
                     where stratC = evalTraversable (rpar `dot` strat) :: Strategy (c a)

-----------------------------------------------------------------------------
-- The new and improved Cluster class

-- Cluster class
--
-- V7: Traversable of clusters (clusters of arbitrary type)
--
class (Traversable c, Monoid a) => Cluster a c where
  cluster    :: Int -> a -> c a
  decluster  :: c a -> a
  lift       :: (a -> b) -> c a -> c b

  decluster = fold   -- c is Foldable (to Monoid a)
  lift = fmap        -- c is a Functor
  -- we require: decluster . cluster n == id

-- instance for lists with chunking
instance Cluster [a] [] where
  cluster = chunk

--
-- generic evalCluster
--
-- 'ClusterStrategy' is a GADT wrapper around the strategy argument to
-- 'evalCluster'. It exposes the cluster constructor 'c' in the type of
-- 'evalCluster'; without this, the type system can't infer 'c' when
-- 'evalCluster' is used.
--
data ClusterStrategy a c where
  ClustStrat :: forall a c . Cluster a c => Strategy a -> ClusterStrategy a c

-- Note that the type annotations in 'evalCluster' and 'parCluster' are
-- forcing the explicit use of 'forall' in the type declarations.
evalCluster :: forall a c . Cluster a c => Int -> ClusterStrategy a c -> Strategy a
evalCluster n (ClustStrat strat) = evalClusterBy (cluster n) decluster (evalTraversable strat :: Strategy (c a))

parCluster :: forall a c . Cluster a c => Int -> ClusterStrategy a c -> Strategy a
parCluster n (ClustStrat strat) = evalCluster n clusterStrat
  where clusterStrat = ClustStrat (rpar `dot` strat) :: ClusterStrategy a c

--
-- generic evalCluster, without GADTs
--

-- wrapper type
data ClusterWith (c :: * -> *) = C

evalCluster0 :: forall a c. Cluster a c => ClusterWith c ->
                 Int -> Strategy a -> Strategy a
evalCluster0 _ n strat =
  \ x -> return (decluster (cluster n x `using` stratC))
  -- evalClusterBy (cluster n) decluster stratC
  where stratC = evalTraversable strat :: Strategy (c a)

parCluster0 :: forall a c. Cluster a c => ClusterWith c ->
                 Int -> Strategy a -> Strategy a
parCluster0 z n strat = evalCluster0 z n (rpar `dot` strat)

-- only for testing compilation of the parMapCluster used in the paper
parMapCluster0 :: forall a b c. Cluster  [b] c =>
                   ClusterWith c -> Strategy [b] -> Int ->
                   (a -> b) -> [a] -> [b]
parMapCluster0 _ strat z f xs =
  map f xs `using`
    evalCluster0 (C::ClusterWith []) z (rpar `dot` strat)

-- generic wrapper type
data ClusterWithX :: (* -> *) -> *

__ = undefined

evalClusterX :: forall a c w . Cluster a c
                 => w c -> Int -> Strategy a -> Strategy a
evalClusterX _ n strat =
  \ x -> return (decluster (cluster n x `using` stratC))
  where stratC = evalTraversable strat :: Strategy (c a)
  -- evalClusterBy (cluster n) decluster stratC

parClusterX :: forall a c w. Cluster a c => w c ->
                 Int -> Strategy a -> Strategy a
parClusterX z n strat = evalClusterX z n (rpar `dot` strat)

-- only for testing compilation of the parMapCluster used in the paper
parMapClusterX :: forall a b c w . Cluster [b] c
                   => w c -> Int -> Strategy b
                   -> (a -> b) -> [a] -> [b]
parMapClusterX _ n strat f xs = map f xs
  `using` evalClusterX (__::w c) n (rpar `dot` evalList strat)

-- --------------------------------------------------------------------------
-- Generic clustering strategies

-- | @'evalClusterBy' cluster decluster strat@ applies @strat@
-- to a clustered input (resulting from applying @cluster@).
-- To meet the requirement on strategies being identities, the @decluster@
-- function must be a left inverse of @cluster@, that is the equation
-- @decluster . cluster == id@ must hold.
evalClusterBy :: (a -> b) -> (b -> a) -> Strategy b -> Strategy a
evalClusterBy cluster decluster strat x
  = return (decluster (cluster x `using` strat))

-- Note: Safety of evalClusterBy depends heavily on its arguments.
--       Maybe it should not be exported but only used internally.

-- | Generic clustering strategy, using an overloaded cluster/decluster pair.
-- n:       cluster size
-- strat:   strategy applied to a cluster
-- NO: liftStrat (or evalMonad) is too lazy to effectively perform computations over the clusters
-- evalCluster :: Cluster c => Int -> Strategy (c a) -> Strategy (c a)
-- evalCluster n strat = evalClusterBy (cluster n) decluster (liftStrat strat)

-- | Generic clustering strategy, using an overloaded cluster/decluster pair and a strategy combinator.
-- n:       cluster size
-- scomb:   strategy combinator lifting strategy from @c a@ to @c (c a)@
-- strat:   strategy applied to a cluster
-- evalClusterComb :: Cluster a c => Int -> (Strategy a -> Strategy (c a)) -> Strategy a -> Strategy a
-- evalClusterComb n scomb strat = evalClusterBy (cluster n) decluster (scomb strat)

-- | Like 'evalCluster' but evaluates each cluster in parallel.
-- n:       cluster size
-- scomb:   strategy combinator lifting strategy from @c a@ to @c (c a)@
-- strat:   strategy applied to a cluster
-- parCluster :: Cluster a c => Int -> Strategy a -> Strategy a
-- parCluster n strat = evalCluster n (rpar `dot` strat)

-- | Like 'evalClusterComb' but evaluates each cluster in parallel.
-- n:       cluster size
-- scomb:   strategy combinator lifting strategy from @c a@ to @c (c a)@
-- strat:   strategy applied to a cluster
-- parClusterComb :: Cluster a c => Int -> (Strategy a -> Strategy (c a)) -> Strategy a -> Strategy a
-- parClusterComb n scomb strat = evalClusterComb n scomb (rpar `dot` strat)

-- evalTraversableCluster :: Cluster a c => Int -> Strategy a -> Strategy a
-- evalTraversableCluster n strat = evalClusterBy (cluster n) decluster (evalTraversable strat :: Strategy (c a))

-- parTraversableCluster :: Cluster a c => Int -> Strategy a -> Strategy a
-- parTraversableCluster n strat = -- evalTraversableCluster n (rpar `dot` strat)
--                                 evalClusterBy (cluster n) decluster (evalTraversable (rpar `dot` strat) :: Strategy (c a)) 
-- --------------------------------------------------------------------------
-- Clustering strategies for lists

-- | 'evalList' on chunks of size n.
evalListChunk :: Int -> Strategy [a] -> Strategy [a]
evalListChunk n strat = evalClusterBy (chunk n) concat (evalList strat)

parListChunk :: Int -> Strategy [a] -> Strategy [a]
parListChunk n strat = evalListChunk n (rpar `dot` strat)


-- | 'evalList' on n slices.
evalListSlice :: Int -> Strategy [a] -> Strategy [a]
evalListSlice n strat = evalClusterBy (slice n) shuffle (evalList strat)

parListSlice :: Int -> Strategy [a] -> Strategy [a]
parListSlice n strat = evalListSlice n (rpar `dot` strat)


-- --------------------------------------------------------------------------
-- Clustering strategies for lazy lists

-- | @'evalBufferChunk' n m strat@ is a rolling buffer holding @n@ chunks
-- (of size @m@), applying @strat@ to each chunk as it goes into the buffer.
evalBufferChunk :: Int -> Int -> Strategy [a] -> Strategy [a]
evalBufferChunk n m strat = evalClusterBy (chunk m) concat (evalBuffer n strat)

parBufferChunk :: Int -> Int -> Strategy [a] -> Strategy [a]
parBufferChunk n m strat = evalBufferChunk n m (rpar `dot` strat)

-- | Like 'evalBufferChunk' but uses the slightly less compositional 
-- 'evalBuffer'' under the hood.
evalBufferChunk' :: Int -> Int -> Strategy [a] -> Strategy [a]
evalBufferChunk' n m strat = evalClusterBy (chunk m) concat (evalBuffer' n strat)

parBufferChunk' :: Int -> Int -> Strategy [a] -> Strategy [a]
parBufferChunk' n m strat = evalBufferChunk' n m (rpar `dot` strat)


-- | @'evalBufferChunk' n m k strat@ is a rolling buffer operating on chunks
-- of slices of input list (to increase granularity and buffer throughput
-- without compromising load balancing).
--
-- @n@:     buffer size (ie. number of chunks in buffer)
-- @m@:     number of slices (and stride between elements)
-- @k@:     chunk size
-- @strat@: strategy applied to chunk before pushing it into buffer
evalBufferSliceChunk :: Int -> Int -> Int -> Strategy [a] -> Strategy [a]
evalBufferSliceChunk n m k strat = evalClusterBy (sliceChunk m k) (shuffleChunk m) (evalBuffer n strat)

parBufferSliceChunk :: Int -> Int -> Int -> Strategy [a] -> Strategy [a]
parBufferSliceChunk n m k strat = evalBufferSliceChunk n m k (rpar `dot` strat)

-- | Like 'evalBufferSliceChunk' but uses the slightly less compositional 
-- 'evalBuffer'' under the hood.
evalBufferSliceChunk' :: Int -> Int -> Int -> Strategy [a] -> Strategy [a]
evalBufferSliceChunk' n m k strat = evalClusterBy (sliceChunk m k) (shuffleChunk m) (evalBuffer' n strat)

parBufferSliceChunk' :: Int -> Int -> Int -> Strategy [a] -> Strategy [a]
parBufferSliceChunk' n m k strat = evalBufferSliceChunk' n m k (rpar `dot` strat)


-- --------------------------------------------------------------------------
-- Clustering functions for lists

-- | @'chunk' n xs@ chops @xs@ into chunks of length (up to) @n@, for @n > 0@.
-- For example,
--
-- > chunk 3 [1..8] == [[1,2,3],[4,5,6],[7,8]]
-- > chunk 8 [1..3] == [[1,2,3]]
--
-- The Prelude function 'concat' is a left inverse of @'chunk' n@, that is
--
-- > concat . chunk n == id
--
chunk :: Int -> [a] -> [[a]]
chunk _n [] = []
chunk n  xs = ys:chunk n zs where (ys,zs) = splitAt n xs

-- Moreover, @'chunk' n@ is a left inverse of 'concat' for lists of lists
-- satisfying the first uniform-length property @ul1@, that is
--
-- > not (ul1 xss) || (chunk (length (head xss)) . concat) xss == xss
-- >   where ul1 []  = True
-- >         ul1 xss = all (== k) $ map length $ init xss
-- >                     where k = length (head xss)
--
-- Note that @'chunk' n xs@ satisfies property @ul1@, for all @n@ and @xs@.


-- | @'slice' n xs@ cuts @xs@ into (up to) @n@ slices, for @n > 0@.
-- For example,
--
-- > slice 3 [1..8] == [[1,4,7],[2,5,8],[3,6]]
-- > slice 8 [1..3] == [[1],[2],[3]]
--
slice :: Int -> [a] -> [[a]]
slice n = transpose . chunk n

-- | Variant of @'slice'@; used to be called @unshuffle@.
-- Note that @slice' n xs@ always yields @n@ slices but trailing slices
-- may be empty. For example,
--
-- > slice' 3 [1..8] == [[1,4,7],[2,5,8],[3,6]]
-- > slice' 8 [1..3] == [[1],[2],[3],[],[],[],[],[]]
--
-- 'slice'' is inefficient for larger values of @n@; not to be exported
slice' :: Int -> [a] -> [[a]]
slice' n xs = [takeEachN (drop i xs) | i <- [0..n-1]]
                where takeEachN []     = []
                      takeEachN (x:xs) = x : takeEachN (drop (n-1) xs)

-- | @'shuffle'@ is a left inverse of @'slice' n@.
shuffle :: [[a]] -> [a]
shuffle = concat . transpose

-- Proof of @shuffle . slice n == id@:
--
-- > (shuffle . slice n) xs
-- >   == (concat . transpose . transpose . chunk n) xs
-- >   == (concat . chunk n) xs
-- >   == xs

-- Moreover, @'slice' n@ is a left inverse of @'shuffle'@ for lists of lists
-- satisfying the second uniform-length property @ul2@, that is
--
-- > not (ul2 xss) || (slice (length xss) . shuffle) xss == xss
-- >   where ul2 []  = True
-- >         ul2 xss = all (== (k-1)) $ dropWhile (== k) $ map length xss
-- >                     where k = length (head xss)
--
-- NB: The list @'slice' n xs@ satisfies property @ul2@, for all @n@ and @xs@.
--     The function @map (chunk m)@ preserves property @ul2@, for all @m@.

-- | Variant of @'shuffle'@; not to be exported.
shuffle' :: [[a]] -> [a]
shuffle' xss = case concat (map (take 1) xss) of
                 []    -> []
                 heads -> heads ++ shuffle' (map (drop 1) xss)

-- | Another variant of @'shuffle'@ (apparently coming from to Eden);
-- not to be exported.
shuffle'' :: [[a]] -> [a]
shuffle'' xxs | and (map null xxs) = []
              | otherwise  = mymaphead xxs ++ shuffle'' (map mytail xxs)
                               where mymaphead []           = []
                                     mymaphead ([]:xxs)     = mymaphead xxs
                                     mymaphead ((x:xs):xxs) = x : mymaphead xxs
                                     mytail [] = []
                                     mytail xs = tail xs


-- | @'sliceChunk' n m xs@ cuts @xs@ into @n@ slices, chops each slice into
-- chunks of size @m@ and shuffles the chunked slices, for @n > 0@ and @m > 0@.
-- (That's a conjugation, wow!). For example,
--
-- > sliceChunk 3 3 [1..8] == [[1,4,7],[2,5,8],[3,6]] == slice 3 [1..8]
-- > sliceChunk 3 2 [1..8] == [[1,4],[2,5],[3,6],[7],[8]]
-- > sliceChunk 3 1 [1..8] == [[1],[2],[3],[4],[5],[6],[7],[8]] == chunk 1 [1..8]
--
sliceChunk :: Int -> Int -> [a] -> [[a]]
sliceChunk n m = shuffle . map (chunk m) . slice n


-- | @'shuffleChunk' n@ is a left inverse of @'sliceChunk' n m@.
shuffleChunk :: Int -> [[a]] -> [a]
shuffleChunk n = shuffle . map concat . slice n

-- Proof of @shuffleChunk n . sliceChunk n m == id@:
--
-- > (shuffleChunk n . sliceChunk n m) xs
-- >   == (shuffle . map concat . slice n . shuffle . map (chunk m) . slice n) xs
-- >   == (shuffle . map concat . map (chunk m) . slice n) xs
-- >   == (shuffle . map (concat . chunk m) . slice n) xs
-- >   == (shuffle . map id . slice n) xs
-- >   == (shuffle . slice n) xs
-- >   == xs
--
-- Note that the second equality holds because @slice n@ is a left inverse of
-- @shuffle@ (since @map (chunk m) $ slice n xs@ satisfies property @ul2@).

-- | Variant of @'shuffleChunk'@; not to be exported.
shuffleChunk' :: Int -> [[a]] -> [a]
shuffleChunk' n xss = case splitAt n xss of
                        ([],  _)   -> []
                        (yss, zss) -> shuffle yss ++ shuffleChunk' n zss
